#!/usr/bin/env python3

from __future__ import annotations
from fabric import Connection, Config
from invoke.exceptions import UnexpectedExit
from pathlib import Path
from io import StringIO
from json import loads, dumps, JSONDecodeError
from os import devnull, PathLike
from sys import stdout, stderr, exit
from contextlib import redirect_stdout
from argparse import ArgumentParser
from ipaddress import IPv4Interface, IPv4Network, IPv6Interface, IPv6Network
from itertools import combinations
from math import comb
from typing import Literal, TypeVar, Callable
from wireguard_tools import WireguardConfig, WireguardPeer, WireguardKey
from attrs import define, field, Attribute
from attrs.validators import optional as validator_optional, in_ as validator_in
from cattrs import structure, unstructure
from cattrs.errors import ClassValidationError

@define
class WovenNode:
    address: str
    gateway: str
    interface: str
    ipv4Ranges: list[str]
    ipv6Ranges: list[str]

T = TypeVar("T", int, float)
def _range_validator(min_value: T, max_value: T) -> Callable[[T], T]:
    def _validate(cls, attribute: Attribute, value: T) -> T:
        if not min_value <= value <= max_value:
            raise ValueError(f"field \"{attribute.name}\" must be between {min_value} and {max_value}")
    return _validate

@define
class WovenConfig:
    ptpIpv4Range: IPv4Network = field(converter = IPv4Network)
    ptpIpv6Range: IPv6Network = field(converter = IPv6Network)
    minPort: int = field(validator = _range_validator(0, 0xFFFF))
    maxPort: int = field(validator = _range_validator(0, 0xFFFF))
    nodes: dict[str, WovenNode]
    ptpIpv4Prefix: int = field(default = 30, validator = _range_validator(0, 32))
    ptpIpv6Prefix: int = field(default = 64, validator = _range_validator(0, 128))
    tunnelPrefix: str = ""
    tunnelSuffix: str = "loop"
    tunnelSeparator: str = "-"
    wireguardDir: Path = field(default = Path("/etc/wireguard"), converter = Path)
    wireguardConfigExt: str = field(default = "conf", converter = lambda x: str(x).lstrip("."))
    table: Literal["auto", "off"] = field(default = "off", validator = validator_optional(validator_in(["auto", "off"])))
    allowedIps: list[str] = field(factory = lambda: ["0.0.0.0/0", "::/0"])
    persistentKeepalive: int = field(default = 20, validator = _range_validator(0, 1 << 31 - 1))

    def _surround(self, val: str) -> str:
        prefix_str = f"{self.tunnelPrefix}{self.tunnelSeparator}" if self.tunnelPrefix else ""
        suffix_str = f"{self.tunnelSeparator}{self.tunnelSuffix}" if self.tunnelSuffix else ""
        return f"{prefix_str}{val}{suffix_str}"
    
    def get_tunnel_name(self, from_id: str, to_id: str) -> str:
        return self._surround(f"{from_id}{self.tunnelSeparator}{to_id}")
    
    def get_config_path(self, tunnel_name: str) -> str:
        return self.wireguardDir / f"{tunnel_name}.{self.wireguardConfigExt}"
    
    @property
    def wireguardConfigGlob(self) -> str:
        return str(self.wireguardDir / f"{self._surround('*')}.{self.wireguardConfigExt}")
    
    @staticmethod
    def from_json_str(config: str) -> WovenConfig:
        return structure(loads(config), WovenConfig)

    @staticmethod
    def load_json_file(path: str | bytes | PathLike) -> WovenConfig:
        return WovenConfig.from_json_str(Path(path).read_text(encoding = "UTF-8"))
    
    def to_json_str(self) -> str:
        return dumps(unstructure(self), indent = 4)
    
    def save_json_file(self, path: str | bytes | PathLike) -> None:
        Path(path).write_text(self.to_json_str(), encoding = "UTF-8")
    
    def validate(self) -> None:
        tunnel_count = comb(len(self.nodes), 2)
        if int(2 ** (self.ptpIpv4Prefix - self.ptpIpv4Range.prefixlen)) < tunnel_count:
            raise ValueError("not enough IPv4 PtP networks to assign")
        if int(2 ** (self.ptpIpv6Prefix - self.ptpIpv6Range.prefixlen)) < tunnel_count:
            raise ValueError("not enough IPv6 PtP networks to assign")
        if int(2 ** (32 - self.ptpIpv4Prefix)) - 2 < 2:
            raise ValueError("not enough IPv4 addresses in each PtP network")
        if int(2 ** (128 - self.ptpIpv6Prefix)) - 2 < 2:
            raise ValueError("not enough IPv6 addresses in each PtP network")
        if self.maxPort - self.minPort < tunnel_count:
            raise ValueError("not enough ports to assign")

    def __attrs_post_init__(self):
        self.validate()
    
    def apply(self, ssh_config = Config(overrides = { "run": { "hide": True } })) -> None:
        ptp_ipv4_network_iter = self.ptpIpv4Range.subnets(new_prefix = self.ptpIpv4Prefix)
        ptp_ipv6_network_iter = self.ptpIpv6Range.subnets(new_prefix = self.ptpIpv6Prefix)
        port_iter = iter(range(self.minPort, self.maxPort))
        
        cs = { id: Connection(node.address, user = "root", config = ssh_config) for id, node in self.nodes.items() }
        
        for id, c in cs.items():
            print(f"stopping services for {id}...", end = " ", flush = True)
            c.run(f"for f in {self.wireguardConfigGlob}; do systemctl stop wg-quick@$(basename $f {self.wireguardConfigExt}).service; done")
            print("done")
            print(f"removing existing configs for {id}...", end = " ", flush = True)
            try:
                c.run(f"rm {self.wireguardConfigGlob}")
            except UnexpectedExit:
                pass
            print("done")
        
        for (id_a, node_a), (id_b, node_b) in combinations(self.nodes.items(), 2):
            print(f"creating configs for {id_a} <-> {id_b} tunnel...", end = " ", flush = True)
            
            try:
                ptp_ipv4_network = next(ptp_ipv4_network_iter)
            except StopIteration:
                raise ValueError("not enough IPv4 PtP networks to assign")
            try:
                ptp_ipv6_network = next(ptp_ipv6_network_iter)
            except StopIteration:
                raise ValueError("not enough IPv6 PtP networks to assign")
            try:
                port = next(port_iter)
            except StopIteration:
                raise ValueError("not enough ports to assign")
            
            ipv4_iter = ptp_ipv4_network.hosts()
            try:
                ipv4_a = next(ipv4_iter)
                ipv4_b = next(ipv4_iter)
            except StopIteration:
                raise ValueError("not enough IPv4 addresses in each PtP network")
            
            ipv6_iter = ptp_ipv6_network.hosts()
            try:
                ipv6_a = next(ipv6_iter)
                ipv6_b = next(ipv6_iter)
            except StopIteration:
                raise ValueError("not enough IPv6 addresses in each PtP network")
            
            key_a = WireguardKey.generate()
            key_a_pub = key_a.public_key()
            
            key_b = WireguardKey.generate()
            key_b_pub = key_b.public_key()
            
            tunnel_name_a = self.get_tunnel_name(id_a, id_b)
            addresses_a = [IPv4Interface(f"{ipv4_a}/{self.ptpIpv4Range.prefixlen}"), IPv6Interface(f"{ipv6_a}/{self.ptpIpv6Range.prefixlen}")]
            preup_a = [f"ip ro replace {node_b.address}/32 dev {node_a.interface} via {node_a.gateway} metric 10 src {node_a.address}"]
            predown_a = [f"ip ro del {node_b.address}/32 dev {node_a.interface} via {node_a.gateway} metric 10 src {node_a.address}"]
            postup_a = [f"ip ro replace {sn} dev {tunnel_name_a} via {ipv4_b} metric 10" for sn in node_b.ipv4Ranges] + [f"ip -6 ro replace {sn} dev {tunnel_name_a} via {ipv6_b} metric 10" for sn in node_b.ipv6Ranges]
            postdown_a = [f"ip ro del {sn} dev {tunnel_name_a} via {ipv4_b} metric 10" for sn in node_b.ipv4Ranges] + [f"ip -6 ro del {sn} dev {tunnel_name_a} via {ipv6_b} metric 10" for sn in node_b.ipv6Ranges]
            
            config_a = WireguardConfig(
                addresses = addresses_a,
                listen_port = port,
                private_key = key_a,
                table = self.table,
                preup = preup_a,
                predown = predown_a,
                postup = postup_a,
                postdown = postdown_a,
                peers = {
                    key_b_pub: WireguardPeer(
                        public_key = key_b_pub,
                        allowed_ips = self.allowedIps,
                        endpoint_host = node_b.address,
                        endpoint_port = port,
                        persistent_keepalive = self.persistentKeepalive
                    )
                }
            )
            
            tunnel_name_b = self.get_tunnel_name(id_b, id_a)
            addresses_b = [IPv4Interface(f"{ipv4_b}/{self.ptpIpv4Range.prefixlen}"), IPv6Interface(f"{ipv6_b}/{self.ptpIpv6Range.prefixlen}")]
            preup_b = [f"ip ro replace {node_a.address}/32 dev {node_b.interface} via {node_b.gateway} metric 10 src {node_b.address}"]
            predown_b = [f"ip ro del {node_a.address}/32 dev {node_b.interface} via {node_b.gateway} metric 10 src {node_b.address}"]
            postup_b = [f"ip ro replace {sn} dev {tunnel_name_b} via {ipv4_a} metric 10" for sn in node_a.ipv4Ranges] + [f"ip -6 ro replace {sn} dev {tunnel_name_b} via {ipv6_a} metric 10" for sn in node_a.ipv6Ranges]
            postdown_b = [f"ip ro del {sn} dev {tunnel_name_b} via {ipv4_a} metric 10" for sn in node_a.ipv4Ranges] + [f"ip -6 ro del {sn} dev {tunnel_name_b} via {ipv6_a} metric 10" for sn in node_a.ipv6Ranges]
            
            config_b = WireguardConfig(
                addresses = addresses_b,
                listen_port = port,
                private_key = key_b,
                table = self.table,
                preup = preup_b,
                predown = predown_b,
                postup = postup_b,
                postdown = postdown_b,
                peers = {
                    key_a_pub: WireguardPeer(
                        public_key = key_a_pub,
                        allowed_ips = self.allowedIps,
                        endpoint_host = node_a.address,
                        endpoint_port = port,
                        persistent_keepalive = self.persistentKeepalive
                    )
                }
            )
            print("done")
            
            print(f"saving {id_a} side of {id_a} <-> {id_b} tunnel...", end = " ", flush = True)
            cs[id_a].put(StringIO(config_a.to_wgconfig(wgquick_format = True)), str(self.get_config_path(tunnel_name_a)))
            print("done")
            print(f"saving {id_b} side of {id_a} <-> {id_b} tunnel...", end = " ", flush = True)
            cs[id_b].put(StringIO(config_b.to_wgconfig(wgquick_format = True)), str(self.get_config_path(tunnel_name_b)))
            print("done")

        for id, c in cs.items():
            print(f"starting services for {id}...", end = " ", flush = True)
            c.run(f"for f in {self.wireguardConfigGlob}; do systemctl start wg-quick@$(basename $f .{self.wireguardConfigExt}).service; done")
            print("done")

def main():
    parser = ArgumentParser("woven")
    parser.add_argument("-q", "--quiet", action = "store_true", help = "decrease output verbosity")
    parser.add_argument("-c", "--config", default = "config.json", help = "the path to the config file")
    parser.add_argument("-v", "--validate", action = "store_true", help = "only validate the config without applying it")
    args = parser.parse_args()
    
    with redirect_stdout(open(devnull, "w") if args.quiet else stdout):
        try:
            config = WovenConfig.load_json_file(args.config)
        except FileNotFoundError:
            print(f"No configuration file found at '{args.config}'", file = stderr)
            exit(1)
        except JSONDecodeError as e:
            print(f"Invalid JSON encountered in configuration file: {e}", file = stderr)
            exit(1)
        except ClassValidationError as e:
            print(f"The following validation errors occurred when loading the configuration file:", file = stderr)
            for e in e.exceptions:
                print(f"{type(e).__name__}: {e}", file = stderr)
            exit(1)
        if args.validate:
            return
        try:
            config.apply()
        except Exception as e:
            print(f"error applying configuration: {e}", file = stderr)
            exit(1)

if __name__ == "__main__":
    main()
